import logging
from dataclasses import dataclass, field
from datetime import datetime
from typing import Optional, Tuple

from omegaconf import MISSING
from pandassta.logging_constants import ISO_STR_FORMAT
from pandassta.sta import DbCredentials, FilterEntry, PhenomenonTimeFilter, Properties

from searegion_detection.queryregion import DbCredentials

log = logging.getLogger(__name__)


@dataclass
class ThingConfig:
    id: int


@dataclass
class SensorThingsAuth:
    username: str
    passphrase: str


@dataclass
class DataApi:
    base_url: str
    things: ThingConfig
    filter: FilterEntry
    auth: SensorThingsAuth


@dataclass
class Range:
    range: Tuple[float, float]


@dataclass
class QcDependentEntry:
    independent: int
    dependent: int
    QC: Range
    dt_tolerance: str


@dataclass
class QcEntry:
    range: Range
    gradient: Range
    zscore: Range


@dataclass
class LocationConfig:
    connection: DbCredentials
    crs: str
    time_window: str
    max_dx_dt: float
    max_ddx_dtdt: float


@dataclass
class ResetConfig:
    overwrite_flags: bool = field(default=False)
    observation_flags: bool = field(default=False)
    feature_flags: bool = field(default=False)
    exit: bool = field(default=False)


@dataclass
class OtherConfig:
    count_observations: bool = field(default=False)


@dataclass
class DateConfig:
    format: str


@dataclass
class TimeConfig:
    start: str
    end: str
    date: DateConfig
    format: str = field(default="%Y-%m-%d %H:%M")
    window: Optional[str] = field(default=None)


@dataclass
class HydraRunConfig:
    dir: str


@dataclass
class HydraConfig:
    run: HydraRunConfig
    verbose: Optional[str] = field(default=None)


@dataclass
class QCconf:
    time: TimeConfig
    hydra: HydraConfig
    data_api: DataApi
    reset: ResetConfig
    other: OtherConfig
    location: LocationConfig
    QC_dependent: list[QcDependentEntry]
    QC: dict[str, QcEntry]
    QC_global: dict[str, QcEntry] = field(default_factory=dict)


def filter_cfg_to_query(filter_cfg: FilterEntry) -> str:
    filter_condition = ""
    if filter_cfg:
        range = filter_cfg.phenomenonTime.range
        format = filter_cfg.phenomenonTime.format

        t0, t1 = [datetime.strptime(str(ti), format) for ti in range]

        filter_condition = (
            f"{Properties.PHENOMENONTIME} gt {t0.strftime(ISO_STR_FORMAT)} and "
            f"{Properties.PHENOMENONTIME} lt {t1.strftime(ISO_STR_FORMAT)}"
        )
    log.debug(f"Configure filter: {filter_condition=}")
    return filter_condition


def get_date_from_string(
    str_in: str, str_format_in: str = "%Y-%m-%d %H:%M", str_format_out: str = "%Y%m%d"
) -> str:
    date_out = datetime.strptime(str(str_in), str_format_in)
    return date_out.strftime(str_format_out)



import pytest
from hydra import compose, initialize
from omegaconf import DictConfig
from pandassta.sta_requests import set_sta_url

from src.df_qc_tools.config import (QCconf, filter_cfg_to_query,
                                                  get_date_from_string)


@pytest.fixture(scope="session")
def cfg() -> DictConfig:
    with initialize(config_path="./conf", version_base="1.2"):
        conf = compose("conf_base.yaml")
    set_sta_url(conf.data_api.base_url)

    return conf


class TestConfig:
    def test_hydra_is_loaded(self):
        print(cfg)
        assert cfg

    def test_filter_cfg_to_query(self, cfg: QCconf):
        out = filter_cfg_to_query(cfg.data_api.filter)
        assert (
            out == "phenomenonTime gt 1002-01-01T00:00:00.000000Z and "
            "phenomenonTime lt 3003-01-01T00:00:00.000000Z"
        )

    def test_get_date_from_string(self):
        # date_o = get_date_from_string("2023-04-01 12:15", "%Y-%m-%d %H:%M", "%Y%m%d")
        date_o = get_date_from_string("2023-04-01 12:15")
        assert date_o == "20230401"



import json
import logging
from copy import deepcopy
from dataclasses import dataclass, field
import operator
from typing import Callable

import geopandas as gpd
import numpy as np
import pandas as pd
import xarray as xr
from scipy import stats
from tqdm import tqdm

from pandassta.df import (
    Df,
    QualityFlags,
    get_acceleration_series,
    get_distance_geopy_series,
    get_velocity_series,
)
from pandassta.df import CAT_TYPE
from pandassta.logging_constants import TQDM_DESC_FORMAT
from pandassta.logging_constants import TQDM_BAR_FORMAT

log = logging.getLogger(__name__)


# not used, kept as reference
# def get_null_mask(df: pd.DataFrame, qc_type: str) -> pd.Series:
#     mask_out = (
#         ~df[[f"qc_{'_'.join([qc_type, i])}" for i in ["min", "max"]]]
#         .isnull()
#         .any(axis=1)
#     )
#     return mask_out


def get_bool_out_of_range(
    df: pd.DataFrame, qc_on: str | tuple, qc_type: str
) -> pd.Series:
    qc_type_min = f"qc_{qc_type}_min"
    qc_type_max = f"qc_{qc_type}_max"

    columns0 = list(df.columns)
    nb_levels = df.columns.nlevels
    for ci in [qc_type_max, qc_type_min]:
        ci_i = ci
        if nb_levels > 1:
            ci_i = (ci,) + ("",) * (nb_levels - 1)
        if ci_i not in columns0:
            columns0 += [ci_i]
    df = df.reindex(columns=columns0)
    mask_max_not_null = ~df[qc_type_max].isnull()
    mask_min_not_null = ~df[qc_type_min].isnull()
    s_bool_out = (df.loc[mask_max_not_null, qc_on] > df.loc[mask_max_not_null, qc_type_max]) | (df.loc[mask_min_not_null, qc_on] < df.loc[mask_min_not_null, qc_type_min])  # type: ignore

    return s_bool_out


def get_bool_null_region(df: pd.DataFrame) -> pd.Series:
    return df[Df.REGION].isnull()


def get_bool_land_region(df: pd.DataFrame) -> pd.Series:
    bool_mainland = df[Df.REGION].str.lower().str.contains("mainland").fillna(False)  # type: ignore
    return bool_mainland


def qc_region(
    df: gpd.GeoDataFrame,
    flag_none: QualityFlags = QualityFlags.PROBABLY_BAD,
    flag_mainland: QualityFlags = QualityFlags.BAD,
) -> gpd.GeoDataFrame:
    df_out = deepcopy(df)

    bool_nan = get_bool_null_region(df_out)
    df_out.loc[bool_nan.index, Df.QC_FLAG] = get_qc_flag_from_bool(
        bool_=bool_nan,
        flag_on_true=flag_none,
    )[Df.QC_FLAG]

    bool_mainland = get_bool_land_region(df_out)
    df_out.loc[bool_mainland.index, Df.QC_FLAG] = get_qc_flag_from_bool(
        bool_=bool_mainland,
        flag_on_true=flag_mainland,
    )[Df.QC_FLAG]

    df_out[Df.QC_FLAG] = df_out[Df.QC_FLAG].astype(CAT_TYPE)  # type: ignore
    log.info(f"Flags set: {df_out.loc[bool_mainland | bool_nan, [Df.QC_FLAG, Df.REGION]].value_counts(dropna=False)}")  # type: ignore
    return df_out


# TODO: refactor, complete df is not needed
def calc_gradient_results(df: pd.DataFrame, groupby: Df) -> pd.DataFrame:
    log.info(f"Start gradient calculations per {groupby}.")
    df_copy = deepcopy(df)

    def grad_function(group):
        nb_row, nb_columns = group.shape
        if nb_row < 2:
            group[Df.GRADIENT] = None
            return group
        g = np.gradient(
            # group.result, group.phenomenonTime.astype("datetime64[s]").astype("int64")
            group.result,
            group.phenomenonTime.astype("datetime64[ns]").astype("int64"),
        )
        group[Df.GRADIENT] = g * 1e9
        return group

    df_tmp = df_copy.sort_values(Df.TIME)
    df_grouped = df_copy.groupby(by=[groupby], group_keys=False)
    df_tmp = df_grouped[[str(Df.RESULT), str(Df.TIME)]].apply(
        grad_function
    )  # casted to string to avoid type error
    df_out = df_copy.join(df_tmp[Df.GRADIENT])
    return df_out


def calc_zscore_results(
    df: pd.DataFrame, groupby: Df, rolling_time_window: str = "60min"
) -> pd.DataFrame:
    df_copy = deepcopy(df)
    def mod_z(df_: pd.DataFrame) -> pd.Series:
        # transformed, _ = stats.yeojohnson(col.values)
        # col[col.columns[0]] = transformed.ravel()
        # df_ = df_.loc[df_.get(Df.QC_FLAG, pd.Series(0, index=df_.index)) <= 2]
        # try:
        # df_ = df_.loc[df_[Df.QC_FLAG] <= 2]
        # except KeyError as e:
        # pass
        roll = df_.sort_values(Df.TIME).rolling(
            rolling_time_window, on=Df.TIME, center=True
        )
        col = df_[Df.RESULT]
        df_["median"] = roll[Df.RESULT].median()
        df_["abs_dev"] = (df_[Df.RESULT] - df_["median"]).abs()
        # med_abs_dev = np.median(np.abs(col - med_col))
        # med_abs_dev = np.abs(roll[Df.RESULT] - med_col).median() # type: ignore
        roll = df_.sort_values(Df.TIME).rolling(
            rolling_time_window, on=Df.TIME, center=True
        )
        df_["med_abs_dev"] = roll["abs_dev"].median()
        # (X-MED)/(1.486*MAD)
        mod_z = (col-df_["median"])/(1.486*df_["med_abs_dev"])
        # mod_z = 0.6745 * ((col - df_["median"]) / df_["med_abs_dev"])
        mod_z.loc[df_["med_abs_dev"] == 0] = (col - df_["median"]) / (
            1.253314 * roll["abs_dev"].mean()
        )
        # return np.abs(mod_z)
        return mod_z

    def _calc_zscore_results(df, groupby):
        # group = df.groupby(by=groupby, group_keys=False)
        # group = df.loc[df.get(Df.QC_FLAG, pd.Series(0, index=df.index).map(QualityFlags)) <= QualityFlags(2)].groupby(
        df_ = df.loc[:]
        df_.loc[df_.get(Df.QC_FLAG, pd.Series(0, index=df_.index).map(QualityFlags)) >= QualityFlags.PROBABLY_BAD, Df.RESULT] = pd.NA
        group = df_.groupby(
            by=groupby, group_keys=False
        )
        # group = df[[Df.TIME, Df.RESULT, groupby]].groupby(by=groupby, group_keys=False)
        # z = group[[Df.TIME, Df.RESULT]].rolling(rolling_time_window=Df.TIME, center=True).apply(stats.zscore)
        z = group[[Df.TIME, Df.RESULT]].apply(mod_z)
        # z = group[[Df.RESULT]].apply(stats.zscore)
        if group.ngroups == 1:
            return z.T
        return z

    df_copy[Df.ZSCORE] = _calc_zscore_results(df_copy, groupby=[Df.DATASTREAM_ID])
    df[Df.ZSCORE] = df_copy[Df.ZSCORE]
    # roll_median = df.loc[:, [Df.RESULT, Df.DATASTREAM_ID, Df.TIME]].sort_values(Df.TIME).rolling(rolling_time_window, on=Df.TIME, center=True).median()
    # df.loc[:, [Df.RESULT, Df.TIME]].sort_values(Df.TIME)
    # df[Df.ZSCORE]
    # roll_median = roll.median()
    return df


def dependent_quantity_merge_asof(
    df: pd.DataFrame, independent: int, dependent: int, dt_tolerance: str
) -> pd.DataFrame:
    df_indep = (
        df.loc[df[Df.DATASTREAM_ID] == independent]
        .sort_values(Df.TIME)
        .set_index(Df.TIME)
    )
    df_dep = (
        df.loc[df[Df.DATASTREAM_ID] == dependent]
        .sort_values(Df.TIME)
        .set_index(Df.TIME)
    )

    # df_merged = pd.merge_asof(df_indep, df_dep, left_index=True, right_index=True, tolerance=pd.Timedelta('0.5s'), suffixes=[f"_{i}" for i in [independent, dependent]])
    df_merged = pd.merge_asof(
        df_dep,
        df_indep,
        left_index=True,
        right_index=True,
        tolerance=pd.Timedelta("0.5s"),
        suffixes=[f"_{i}" for i in [dependent, independent]],
    )
    df_merged = pd.DataFrame(
        df_merged.values,
        index=df_merged.index,
        columns=df_merged.columns.str.rsplit("_", expand=True, n=1),
    )
    # df_merged = df_type_conversions(df_merged)

    return df_merged


def dependent_quantity_pivot(
    df: pd.DataFrame, independent: int, dependent: int, dt_tolerance: str
) -> pd.DataFrame:
    # merge_asof is used, but creates a pivot-like table
    df_merged = dependent_quantity_merge_asof(
        df, independent=independent, dependent=dependent, dt_tolerance=dt_tolerance
    )
    return df_merged


def strip_df_to_minimal_required_dependent_quantity(df, independent, dependent):
    df_out = deepcopy(
        df.reset_index().loc[
            df[Df.DATASTREAM_ID].isin([independent, dependent]),
            [
                Df.TIME,
                Df.DATASTREAM_ID,
                Df.RESULT,
                Df.QC_FLAG,
                Df.OBSERVATION_TYPE,
                Df.IOT_ID,
            ],
        ]
    )
    return df_out


def get_bool_flagged_dependent_quantity(
    df: pd.DataFrame, independent: int, dependent: int, dt_tolerance: str
) -> pd.Series:
    df_tmp = strip_df_to_minimal_required_dependent_quantity(
        df, independent=independent, dependent=dependent
    )

    df_pivot = dependent_quantity_pivot(
        df_tmp, independent=independent, dependent=dependent, dt_tolerance=dt_tolerance
    )

    mask = ~df_pivot[Df.QC_FLAG, str(independent)].isin(
        [QualityFlags.NO_QUALITY_CONTROL, QualityFlags.GOOD]
    )
    bool_ = df[Df.IOT_ID].isin(df_pivot.loc[mask, (Df.IOT_ID, str(dependent))].values)  # type: ignore
    return bool_


def drop_duplicates_df_unpivot(df_unpivot: pd.DataFrame, independent: int, dependent: int) -> pd.DataFrame:
    if df_unpivot.index.has_duplicates:
        datastreams_with_duplicate_idx = list(df_unpivot[df_unpivot.index.duplicated(keep='first')][Df.DATASTREAM_ID].unique())
        log.warning(f"Duplicated {Df.IOT_ID} found for {Df.DATASTREAM_ID} {datastreams_with_duplicate_idx}.")
        log.warning(f"This might be due to duplicate entries for {set([independent, dependent]).difference(set(datastreams_with_duplicate_idx))}.")
        log.warning(f"Duplicates will be removed.")
        df_unpivot = df_unpivot[~df_unpivot.index.duplicated(keep='first')]

    return df_unpivot


def qc_dependent_quantity_base(
    df: pd.DataFrame,
    independent: int,
    dependent: int,
    dt_tolerance: str,
    flag_when_missing: QualityFlags | None = QualityFlags.BAD,
) -> pd.Series:
    df_tmp = strip_df_to_minimal_required_dependent_quantity(
        df, independent=independent, dependent=dependent
    )

    df_pivot = dependent_quantity_pivot(
        df_tmp, independent=independent, dependent=dependent, dt_tolerance=dt_tolerance
    )

    mask = ~df_pivot[Df.QC_FLAG, str(independent)].isin(
        [QualityFlags.NO_QUALITY_CONTROL, QualityFlags.GOOD]
    )
    df_pivot.loc[mask, (Df.QC_FLAG, str(dependent))] = df_pivot[mask][
        (Df.QC_FLAG, str(independent))
    ]

    df_unpivot = df_pivot.loc[mask].stack(future_stack=True).dropna(subset=Df.IOT_ID).reset_index().set_index(Df.IOT_ID)  # type: ignore
    df_unpivot = drop_duplicates_df_unpivot(df_unpivot=df_unpivot, independent=independent, dependent=dependent)
    
    # df_unpivot = df_pivot.loc[mask].stack().reset_index().set_index(Df.IOT_ID)
    df = df.set_index(Df.IOT_ID)
    # TODO: refactor
    mask_unpivot_notnan = ~df_unpivot[Df.QC_FLAG].isna()
    idx_unpivot_notnan = df_unpivot.loc[mask_unpivot_notnan, Df.QC_FLAG].index.astype(
        int
    )  # the conversion to int is needed because nan is float, and this column is set to float for some reason
    idx_unpivot_nan = df_unpivot.loc[~mask_unpivot_notnan, Df.QC_FLAG].index.astype(int)
    df.loc[idx_unpivot_notnan, Df.QC_FLAG] = df_unpivot.loc[
        idx_unpivot_notnan, Df.QC_FLAG
    ]
    s_out = df.loc[idx_unpivot_notnan, Df.QC_FLAG]
    if flag_when_missing:
        df.loc[idx_unpivot_nan, Df.QC_FLAG] = flag_when_missing  # type: ignore
        s_out = df.loc[idx_unpivot_notnan.union(idx_unpivot_nan), Df.QC_FLAG]
    return s_out  # type: ignore


def qc_dependent_quantity_secondary(
    df: pd.DataFrame,
    independent: int,
    dependent: int,
    range_: tuple[float, float],
    dt_tolerance: str,
) -> pd.Series:
    df_tmp = strip_df_to_minimal_required_dependent_quantity(
        df, independent=independent, dependent=dependent
    )

    df_pivot = dependent_quantity_pivot(
        df_tmp, dependent=dependent, independent=independent, dt_tolerance=dt_tolerance
    )

    df_pivot[["qc_drange_min", "qc_drange_max"]] = range_
    bool_qc = get_bool_out_of_range(
        df_pivot, (Df.RESULT, str(independent)), qc_type="drange"
    )
    df_pivot.loc[bool_qc, (Df.QC_FLAG, str(dependent))] = QualityFlags.BAD  # type: ignore Don"t know how to fix this

    df_pivot = df_pivot.drop(["qc_drange_min", "qc_drange_max"], axis=1, level=0)
    df_unpivot = df_pivot.stack(future_stack=True).dropna(subset=Df.IOT_ID).reset_index().set_index(Df.IOT_ID)  # type: ignore

    df_unpivot = drop_duplicates_df_unpivot(df_unpivot=df_unpivot, independent=independent, dependent=dependent)

    # df_unpivot = df_pivot.stack(future_stack=True).reset_index().set_index(Df.IOT_ID)  # type: ignore
    df = df.set_index(Df.IOT_ID)
    df.loc[df_unpivot.index, Df.QC_FLAG] = df_unpivot[Df.QC_FLAG]
    s_out = df.loc[df_unpivot.index, Df.QC_FLAG]
    return s_out  # type: ignore


def get_qc_flag_from_bool(
    bool_: pd.Series,
    flag_on_true: QualityFlags,
    flag_on_false: QualityFlags | None = None,
) -> pd.Series:
    qc_flag_series = pd.Series(flag_on_true, index=bool_.index, dtype=CAT_TYPE).loc[  # type: ignore
        bool_
    ]
    if flag_on_false:
        qc_flag_series = pd.concat(
            [
                qc_flag_series,
                pd.Series(flag_on_false, index=bool_.index, dtype=CAT_TYPE).loc[~bool_],  # type: ignore
            ]
        )
    return qc_flag_series


# TODO: refactor
def get_bool_spacial_outlier_compared_to_median(
    df: gpd.GeoDataFrame, max_dx_dt: float, time_window: str
) -> pd.Series:
    def delta(x):
        return np.max(x) - np.min(x)

    log.info("Start calculating spacial outliers.")
    df_time_sorted = df.sort_values(Df.TIME)
    # df_time_sorted["dt"] = df_time_sorted.loc[:, Df.TIME].dt.total_seconds()
    # df_time_sorted["dt"] = df_time_sorted[Df.TIME].diff().fillna(pd.to_timedelta("0")).dt.total_seconds()  # type: ignore
    df_time_sorted["dt"] = (df_time_sorted[Df.TIME] - df_time_sorted[Df.TIME].min()).dt.total_seconds()  # type: ignore

    bool_series_lat_eq_long = df_time_sorted[Df.LAT] == df_time_sorted[Df.LONG]
    bool_series_lat_or_long_zero = (df_time_sorted[Df.LAT] == 0) | (
        df_time_sorted[Df.LONG] == 0
    )
    entries_excluded_from_calculations = (
        bool_series_lat_eq_long | bool_series_lat_or_long_zero
    )

    log.debug(
        f"{bool_series_lat_eq_long.value_counts(dropna=False)=} (excluded from median calculations)"
        f"{bool_series_lat_or_long_zero.value_counts(dropna=False)=} (excluded from median calculations)"
        f"{entries_excluded_from_calculations.value_counts(dropna=False)=} (excluded from median calculations)"
    )

    tqdm.pandas(
        total=df.shape[0],
        desc=TQDM_DESC_FORMAT.format("Rolling median"),
        bar_format=TQDM_BAR_FORMAT,
    )
    log.debug("Start rolling median calculations.")
    rolling_median = (
        df_time_sorted.loc[
            ~entries_excluded_from_calculations, [Df.TIME, Df.LONG, Df.LAT]
        ]
        .sort_values(Df.TIME)
        .rolling(time_window, on=Df.TIME, center=True)
        .median()
        # .progress_apply(np.median)  # type: ignore
    )

    tqdm.pandas(
        total=df.shape[0],
        desc=TQDM_DESC_FORMAT.format("Rolling time"),
        bar_format=TQDM_BAR_FORMAT,
    )
    log.debug("Start rolling time calculations.")
    # calculates the time delta in each windows
    # rolling_time = (
    #     df_time_sorted.loc[:, [Df.TIME, "dt"]]
    #     .sort_values(Df.TIME)
    #     .rolling(time_window, on=Df.TIME, center=True)
    #     .apply(delta)
    # )
    # rolling_time = (rolling_t.max() - rolling_t.min()).dt
    rolling_t = df_time_sorted.loc[:, [Df.TIME, "dt"]].rolling(
        time_window, on=Df.TIME, center=True
    )
    rolling_time = rolling_t.max() - rolling_t.min()

    rolling_time[Df.TIME] = df_time_sorted[Df.TIME]

    rolling_median = rolling_median.reindex(index=rolling_time.index, fill_value=None)
    rolling_median.loc[entries_excluded_from_calculations, Df.TIME] = rolling_time.loc[
        entries_excluded_from_calculations, Df.TIME
    ]
    rolling_median = rolling_median.ffill().bfill()

    ref_point = gpd.GeoDataFrame(  # type: ignore
        rolling_median,
        geometry=gpd.points_from_xy(
            rolling_median.loc[:, Df.LONG],
            rolling_median.loc[:, Df.LAT],
            crs="EPSG:4326",
        ),
    )
    ref_point["dt"] = rolling_time["dt"]
    df_time_sorted["geo_ref"] = ref_point.geometry
    df_time_sorted = gpd.GeoDataFrame(df_time_sorted)
    distance = get_distance_geopy_series(
        df_time_sorted, column1="geometry", column2="geo_ref"
    )
    bool_series = distance.values > (ref_point["dt"] * max_dx_dt).values  # type: ignore
    bool_out = pd.Series(bool_series, index=df_time_sorted.index)
    return bool_out


def get_bool_exceed_max_velocity(
    df: gpd.GeoDataFrame,
    max_velocity: float,
    velocity_series: pd.Series | None = None,
    dt_series: pd.Series | None = None,
) -> pd.Series:
    log.info("Calculating velocity outliers.")
    if velocity_series is None:
        dt_series, velocity_series = get_velocity_series(df, return_dt=True)
    if dt_series is None:
        bool_velocity = (
            velocity_series.abs() > max_velocity
        )  # *dt_series/(dt_series+pd.Timedelta())
    else:
        dt_plus_accuracy = (
            dt_series
            + ((dt_series.dropna() - dt_series.dropna().astype(int)) == 0).astype(int)
            * 1.0
        )
        bool_velocity = (
            velocity_series.abs()
            > (max_velocity / dt_series * dt_plus_accuracy)
            .rename("velocity")
            .loc[velocity_series.index]
        )
    # df["idx_"] = df.index
    # df_tmp = df.set_index(Df.FEATURE_ID)
    # df_tmp["bool_velocity"] = bool_velocity
    # bool_out = df_tmp.set_index("idx_")["bool_velocity"]
    # return bool_out
    return bool_velocity.loc[~velocity_series.isnull()]


def get_bool_exceed_max_acceleration(
    df: gpd.GeoDataFrame,
    max_acceleration: float,
    acceleration_series: pd.Series | None = None,
    dt_series: pd.Series | None = None,
) -> pd.Series:
    log.info("Calculating acceleration outliers.")
    if acceleration_series is None:
        dt_series, acceleration_series = get_acceleration_series(df, return_dt=True)
    if dt_series is None:
        bool_acceleration = acceleration_series.abs() > max_acceleration
    else:
        dt_plus_accuracy = (
            dt_series
            + ((dt_series.dropna() - dt_series.dropna().astype(int)) == 0).astype(int)
            * 1.0
        )
        bool_acceleration = (
            acceleration_series.abs()
            > (max_acceleration / dt_series**2 * dt_plus_accuracy**2)
            .rename("acceleration")
            .loc[acceleration_series.index]
        )

    # df["idx_"] = df.index
    # df_tmp = df.set_index(Df.FEATURE_ID)
    # df_tmp["bool_acceleration"] = bool_acceleration
    # bool_out = df_tmp.set_index("idx_")["bool_acceleration"]
    # return bool_out
    return bool_acceleration.loc[~acceleration_series.isnull()]


def get_depth_from_etop(lat: pd.Series, lon: pd.Series):
    datasetx = xr.open_dataset("./resources/ETOPO_2022_v1_60s_N90W180_bed.nc")

    coords_dataarray = xr.DataArray(
        list(zip(lat, lon)), dims=["points", "coords"], name="coords"
    )
    z_values = (
        datasetx["z"]
        .sel(lat=coords_dataarray[:, 0], lon=coords_dataarray[:, 1], method="nearest")
        .values
    )

    return z_values


def get_bool_depth_below_threshold(df: pd.DataFrame, threshold: float) -> pd.Series:
    mask_is_none = df[Df.REGION].isnull()  # type: ignore
    df_coords_none_unique = df.loc[mask_is_none, [Df.LONG, Df.LAT]]  # type: ignore
    bool_depth = (
        get_depth_from_etop(
            lat=df_coords_none_unique[Df.LAT],  # type: ignore
            lon=df_coords_none_unique[Df.LONG],  # type: ignore
        )
        < threshold
    )
    bool_out = pd.Series(bool_depth, index=df.loc[mask_is_none].index)  # type: ignore
    return bool_out


def get_bool_depth_above_treshold(df: pd.DataFrame, threshold: float) -> pd.Series:
    bool_out = get_bool_depth_below_threshold(df, threshold=threshold)
    return ~bool_out


@dataclass
class QCFlagConfig:
    label: str
    bool_function: Callable
    bool_merge_function: Callable
    flag_on_true: QualityFlags
    flag_on_false: QualityFlags | None = None
    flag_on_nan: QualityFlags | None = None
    bool_series: pd.Series = field(default_factory=pd.Series)
    series_out: pd.Series = field(default_factory=pd.Series)

    def execute(self, df: pd.DataFrame | gpd.GeoDataFrame):
        self.bool_series = self.bool_function(df)
        series_out = (
            df[Df.QC_FLAG].combine(  # type: ignore
                get_qc_flag_from_bool(
                    bool_=self.bool_series,
                    flag_on_true=self.flag_on_true,
                    flag_on_false=self.flag_on_false,
                ),  # type: ignore
                self.bool_merge_function,
                fill_value=self.flag_on_nan,  # type: ignore
            )
        ).astype(CAT_TYPE)
        log.info(f"Execution {self.label} qc result: {self.bool_series.sum()} True")
        self.series_out = series_out  # type: ignore
        return self.series_out


def combine_dicts(a, b, op=operator.add):
    return a | b | dict([(k, op(a[k], b[k])) for k in set(b) & set(a)])


def merge_json_str(jsonstr1: str, jsonstr2: str) -> str:
    d1 = json.loads(jsonstr1)
    d2 = json.loads(jsonstr2)
    # d_out = {key: value for (key, value) in (d1.items() + d2.items())}
    d_out = combine_dicts(d1, d2)

    jsonstr_out = json.dumps(d_out)
    return jsonstr_out


def update_flag_history_series(flag_history_series, flag_config: QCFlagConfig):
    hist_tmp = pd.Series(
        json.dumps({str(flag_config.flag_on_true): [flag_config.label]}),
        index=flag_config.bool_series.loc[flag_config.bool_series].index,
    )

    history_series = hist_tmp.combine(
        flag_history_series, merge_json_str, fill_value=json.dumps({})
    )
    return history_series


FEATURES_BODY_TEMPLATE = '{"properties": {"resultQuality": "{value}"}}'



from copy import deepcopy
from itertools import cycle, islice, product
from operator import add
from typing import Sequence

import geopandas as gpd
import geopy.distance as gp_distance
import numpy as np
import pandas as pd
import pandas.testing as pdt
import pytest
from geopy import Point as gp_point
from pandassta.df import CAT_TYPE, Df, QualityFlags, df_type_conversions

from src.df_qc_tools.qc import (
    calc_gradient_results,
    calc_zscore_results,
    combine_dicts,
    get_bool_exceed_max_acceleration,
    get_bool_exceed_max_velocity,
    get_bool_land_region,
    get_bool_null_region,
    get_bool_out_of_range,
    get_bool_spacial_outlier_compared_to_median,
    get_qc_flag_from_bool,
    qc_dependent_quantity_base,
    qc_dependent_quantity_secondary,
)
from searegion_detection.queryregion import build_points_query


@pytest.fixture
def df_velocity_acceleration() -> gpd.GeoDataFrame:
    df_t = pd.read_csv("./tests/resources/data_velocity_acc.csv", header=0)
    df_t[Df.TIME] = pd.to_timedelta(df_t["Time (s)"], "s") + pd.Timestamp("now")

    p0 = gp_point(longitude=3.1840709669760137, latitude=51.37115902107277)
    for index, row_i in df_t.iterrows():
        di = gp_distance.distance(meters=row_i["Distance (m)"])
        pi = di.destination(point=p0, bearing=row_i["Heading (degrees)"])

        df_t.loc[index, [Df.LONG, Df.LAT]] = pi.longitude, pi.latitude  # type: ignore
        p0 = pi

    df_t = df_t.drop(columns=["Time (s)", "Distance (m)", "Heading (degrees)"])
    df_t = gpd.GeoDataFrame(df_t, geometry=gpd.points_from_xy(df_t[Df.LONG], df_t[Df.LAT], crs="EPSG:4326"))  # type: ignore
    return df_t


@pytest.fixture
def points_testing() -> Sequence[Sequence[float]]:
    points = [
        (3.1840709669760137, 51.37115902107277),  # in front of Zeebrugge (North Sea)
        (3.2063475368848096, 51.34661128136423),  # harbor Zeebrugge
        (3.2270907198892886, 51.21143804531792),  # Brugge
        (3.5553432400042584, 56.169369769668116),  # North Sea
    ]

    return points


base_list_region: list = [
    "NORTH SEA",
    "MAINLAND EUROPE",
    "MAINLAND random",
    None,
    np.nan,
]


@pytest.fixture
def df_testing() -> gpd.GeoDataFrame:
    results_factor: float = 2.345

    base_list_phenomenonTime: list[pd.Timestamp] = list(
        pd.Timestamp("now")
        + pd.timedelta_range(
            start=0, periods=len(base_list_region), freq="s", unit="s"  # type: ignore
        )
    )
    base_results: list[float] = [
        fi * results_factor for fi in range(len(base_list_region))
    ]
    qc_ref_base: list[QualityFlags | float] = [
        QualityFlags.NO_QUALITY_CONTROL,
        QualityFlags.BAD,
        QualityFlags.BAD,
        QualityFlags.PROBABLY_BAD,
        QualityFlags.PROBABLY_BAD,
    ]
    ds_id_type_dict: dict[int, str] = {
        0: "salinity",
        1: "Water flow in the scientific seawater circuit",
        2: "seabed depth",
        3: "random1",
        4: "random2",
    }
    # base_observation_type: list[str] = [
    #     "salinity",
    #     "Water flow in the scientific seawater circuit",
    #     "seabed depth",
    #     "random1",
    #     "random2",
    # ]

    datastream_id_series: "pd.Series[int]" = pd.Series(
        list(
            sum(  # to convert list of tuples to flat list
                zip(*([list(range(MULTIPL_FACTOR))] * len(base_list_region))), ()
            )
        ),
        dtype=int,
    )
    df_out = gpd.GeoDataFrame(
        {
            Df.IOT_ID: pd.Series(range(len(qc_ref_base) * MULTIPL_FACTOR), dtype=int),
            Df.REGION: pd.Series(base_list_region * MULTIPL_FACTOR, dtype="string"),
            Df.QC_FLAG
            + "_ref": pd.Series(qc_ref_base * MULTIPL_FACTOR, dtype=CAT_TYPE),
            Df.QC_FLAG: pd.Series(
                QualityFlags.NO_QUALITY_CONTROL,  # type: ignore
                index=datastream_id_series.index,
                dtype=CAT_TYPE,
            ),  # type: ignore
            Df.DATASTREAM_ID: datastream_id_series,
            Df.TIME: pd.Series(base_list_phenomenonTime * MULTIPL_FACTOR),
            Df.RESULT: pd.Series(
                map(
                    add,
                    base_results * MULTIPL_FACTOR,
                    [i * 10 for i in datastream_id_series.to_list()],
                ),  # type: ignore
                dtype=float,
            ),  # type: ignore
            Df.OBSERVATION_TYPE: datastream_id_series.apply(
                lambda x: ds_id_type_dict[x]

            ),
        }
    )
    return df_out


@pytest.fixture
def df_outliers() -> pd.DataFrame:
    df = pd.read_csv("./tests/resources/df_outliers.csv", header=0, index_col=0)
    df.columns = [Df(ci) for ci in df.columns]
    df[Df.QC_FLAG] = df[Df.QC_FLAG].apply(QualityFlags).astype(CAT_TYPE)  # type: ignore
    df[Df.TIME] = pd.to_datetime(df[Df.TIME])
    return df


MULTIPL_FACTOR: int = 2
LENGTH: int = len(base_list_region) * MULTIPL_FACTOR
LENGTH_SLICE: int = len(base_list_region)


def test_build_points_query(points_testing):
    q: str = build_points_query(points_testing)
    substr_count = q.count("ST_SetSRID(ST_MakePoint(")
    srid_count = q.count("), 4326))")
    assert substr_count == len(points_testing) and srid_count == len(points_testing)


@pytest.mark.skip(reason="bit ridiculous to test")
def test_build_query_seavox():
    assert 0


@pytest.mark.skip(reason="is response dependency")
def test_location_north_sea():
    assert 0


@pytest.mark.skip(reason="is response dependency")
def test_location_mainland_eu():
    assert 0


def test_qc_region_to_flag(df_testing):
    # pandas does strange things with type hints
    # df_out = qc_region(df_testing)
    df_out = deepcopy(df_testing)
    bool_nan = get_bool_null_region(df_out)
    df_out[Df.QC_FLAG] = (
        df_out[Df.QC_FLAG]
        .combine(
            get_qc_flag_from_bool(
                bool_=bool_nan,
                flag_on_true=QualityFlags.PROBABLY_BAD,
            ),
            max,
            fill_value=QualityFlags.NO_QUALITY_CONTROL,
        )
        .astype(CAT_TYPE)
    )

    bool_mainland = get_bool_land_region(df_out)
    df_out[Df.QC_FLAG] = (
        df_out[Df.QC_FLAG]
        .combine(
            get_qc_flag_from_bool(
                bool_=bool_mainland,
                flag_on_true=QualityFlags.BAD,
            ),
            max,
            fill_value=QualityFlags.NO_QUALITY_CONTROL,
        )
        .astype(CAT_TYPE)
    )

    pdt.assert_series_equal(
        df_out.loc[:, Df.QC_FLAG],
        df_out.loc[:, Df.QC_FLAG + "_ref"],
        check_names=False,
    )


def test_qc_region_to_flag_set_on_false(df_testing):
    # pandas does strange things with type hints
    # df_out = qc_region(df_testing)
    df_out = deepcopy(df_testing)
    bool_nan = get_bool_null_region(df_out)
    df_out[Df.QC_FLAG] = (
        df_out[Df.QC_FLAG]
        .combine(
            get_qc_flag_from_bool(
                bool_=bool_nan,
                flag_on_true=QualityFlags.PROBABLY_BAD,
                flag_on_false=QualityFlags.GOOD,
            ),
            max,
            fill_value=QualityFlags.NO_QUALITY_CONTROL,
        )
        .astype(CAT_TYPE)
    )

    bool_mainland = get_bool_land_region(df_out)
    df_out[Df.QC_FLAG] = (
        df_out[Df.QC_FLAG]
        .combine(
            get_qc_flag_from_bool(
                bool_=bool_mainland,
                flag_on_true=QualityFlags.BAD,
            ),
            max,
            fill_value=QualityFlags.NO_QUALITY_CONTROL,
        )
        .astype(CAT_TYPE)
    )

    df_out.loc[
        df_out[Df.QC_FLAG + "_ref"] == QualityFlags.NO_QUALITY_CONTROL,
        Df.QC_FLAG + "_ref",
    ] = QualityFlags.GOOD

    pdt.assert_series_equal(
        df_out.loc[:, Df.QC_FLAG],
        df_out.loc[:, Df.QC_FLAG + "_ref"],
        check_names=False,
    )


@pytest.mark.parametrize(
    "idx,dx,columns",
    [
        ([1, 4], 1, [Df.LONG]),
        ([3, 4], 1, [Df.LAT]),
        ([3, 4], -0.1, [Df.LONG]),
        ([3, 4], -0.1, [Df.LAT, Df.LONG]),
        ([3, 6], -1, [Df.LAT]),
    ],
)
def test_location_outlier(df_testing, idx, dx, columns):
    df_testing[Df.LONG] = df_testing.index * 0.001 + 50.0
    df_testing[Df.LAT] = df_testing.index * 0.001 + 20.0

    for idx_i, col_i in product(idx, columns):
        df_testing.iloc[idx_i, df_testing.columns.get_loc(col_i)] -= dx

    # df_testing.set_geometry("geometry")
    df_testing = df_testing.set_geometry(
        gpd.points_from_xy(df_testing[Df.LONG], df_testing[Df.LAT], crs="EPSG:4326")
    )

    res = get_bool_spacial_outlier_compared_to_median(
        df_testing, max_dx_dt=300.0, time_window="5min"
    )
    mask = np.ma.masked_array(res, mask=res)
    assert all(res[idx]) and ~mask.any() and (sum(res) == len(idx))


@pytest.mark.parametrize(
    "idx,column",
    [
        ([1], Df.LONG),
        ([1, 4], Df.LONG),
        ([1, 4], Df.LAT),
        ([4], Df.LAT),
        ([3, 4], Df.LAT),
        ([3, 4], Df.LONG),
        ([3], Df.LONG),
        ([3, 6], Df.LONG),
        ([6], Df.LONG),
    ],
)
def test_location_outlier_eq(df_testing, idx, column):
    df_testing[Df.LONG] = df_testing.index * 0.001 + 50.0
    df_testing[Df.LAT] = df_testing.index * 0.001 + 20.0

    col1 = column
    set_column_tmp = set([Df.LONG, Df.LAT])
    set_column_tmp.remove(column)
    col2 = set_column_tmp.pop()
    for idx_i in idx:
        df_testing.iloc[idx_i, df_testing.columns.get_loc(col1)] = df_testing.iloc[
            idx_i, df_testing.columns.get_loc(col2)
        ]

    # df_testing.set_geometry("geometry")
    df_testing = df_testing.set_geometry(
        gpd.points_from_xy(df_testing[Df.LONG], df_testing[Df.LAT], crs="EPSG:4326")
    )

    res = get_bool_spacial_outlier_compared_to_median(
        df_testing, max_dx_dt=300.0, time_window="5min"
    )
    mask = np.ma.masked_array(res, mask=res)
    assert all(res[idx]) and ~mask.any() and (sum(res) == len(idx))


@pytest.mark.parametrize(
    "idx,columns",
    [
        ([1, 4], [Df.LONG]),
        ([1, 4], [Df.LAT]),
        ([1, 4], [Df.LAT, Df.LONG]),
        ([3, 4], [Df.LAT]),
        ([3, 4], [Df.LONG]),
        ([3, 6], [Df.LONG]),
        ([3, 6], [Df.LONG, Df.LAT]),
    ],
)
def test_location_outlier_zero(df_testing, idx, columns):
    df_testing[Df.LONG] = df_testing.index * 0.001 + 50.0
    df_testing[Df.LAT] = df_testing.index * 0.001 + 20.0

    for idx_i, col_i in product(idx, columns):
        df_testing.iloc[idx_i, df_testing.columns.get_loc(col_i)] = 0

    # df_testing.set_geometry("geometry")
    df_testing = df_testing.set_geometry(
        gpd.points_from_xy(df_testing[Df.LONG], df_testing[Df.LAT], crs="EPSG:4326")
    )

    res = get_bool_spacial_outlier_compared_to_median(
        df_testing, max_dx_dt=300.0, time_window="5min"
    )
    mask = np.ma.masked_array(res, mask=res)
    assert all(res[idx]) and ~mask.any() and (sum(res) == len(idx))


@pytest.mark.parametrize(
    "idx,dx,columns",
    [
        ([1, 4], 1, [Df.LONG]),
        ([3, 4], 1, [Df.LAT]),
        ([3, 4], -0.1, [Df.LONG]),
        ([3, 4], -0.1, [Df.LAT, Df.LONG]),
        ([3, 6], -1, [Df.LAT]),
    ],
)
def test_location_outlier_long_eq_lat(df_testing, idx, dx, columns):
    df_testing[Df.LONG] = df_testing.index * 0.001 + 50.0
    df_testing[Df.LAT] = df_testing.index * 0.001 + 20.0

    for idx_i, col_i in product(idx, columns):
        df_testing.iloc[idx_i, df_testing.columns.get_loc(col_i)] -= dx
        other_column = [Df.LAT, Df.LONG][col_i == Df.LAT]
        changed_to_value = df_testing.iloc[idx_i, df_testing.columns.get_loc(col_i)]
        df_testing.iloc[idx_i, df_testing.columns.get_loc(other_column)] = (
            changed_to_value
        )

    df_testing = df_testing.set_geometry(
        gpd.points_from_xy(df_testing[Df.LONG], df_testing[Df.LAT], crs="EPSG:4326")
    )

    res = get_bool_spacial_outlier_compared_to_median(
        df_testing, max_dx_dt=300, time_window="5min"
    )
    mask = np.ma.masked_array(res, mask=res)
    assert all(res[idx]) and ~mask.any() and (sum(res) == len(idx))


def test_exceed_max_velocity(df_velocity_acceleration):
    res = get_bool_exceed_max_velocity(df_velocity_acceleration, max_velocity=90)
    assert all(~res)


def test_exceed_max_velocity_2(df_velocity_acceleration):
    res = get_bool_exceed_max_velocity(df_velocity_acceleration, max_velocity=0)
    assert all(res)


def test_exceed_max_velocity_3(df_velocity_acceleration):
    bool_ref = get_bool_exceed_max_velocity(df_velocity_acceleration, max_velocity=1e12)
    bool_ref.loc[3] = True  # type: ignore
    df_velocity_acceleration.loc[4, Df.TIME] = df_velocity_acceleration.loc[3, Df.TIME] + pd.Timedelta(nanoseconds=1)  # type: ignore
    res = get_bool_exceed_max_velocity(df_velocity_acceleration, max_velocity=90)

    pdt.assert_series_equal(res, bool_ref, check_names=False)


@pytest.mark.parametrize(
    "idx0",
    [
        1,
        2,
        3,
        4,
        5,
        6,
    ],
)
def test_exceed_max_acceleration(df_velocity_acceleration, idx0):
    bool_ref = get_bool_exceed_max_acceleration(
        df=df_velocity_acceleration, max_acceleration=1e12
    )
    bool_ref.loc[[idx0 - 1, idx0]] = True
    df_velocity_acceleration.loc[idx0 + 1, Df.TIME] = df_velocity_acceleration.loc[idx0, Df.TIME] + pd.Timedelta(nanoseconds=1)  # type: ignore
    res = get_bool_exceed_max_acceleration(
        df=df_velocity_acceleration, max_acceleration=25
    )

    pdt.assert_series_equal(res, bool_ref, check_names=False)


@pytest.mark.parametrize(
    "result,ref",
    [
        (np.zeros(LENGTH), np.zeros(LENGTH)),
        (range(0, LENGTH, 1), np.ones(LENGTH)),
        (range(LENGTH, 0, -1), np.ones(LENGTH) * -1),
    ],
)
def test_qc_gradient_calc_basic(df_testing, result, ref):
    df_testing[Df.TIME] = pd.Timestamp("now") + pd.timedelta_range(
        start=0, periods=df_testing.shape[0], freq="s", unit="s"  # type: ignore
    )
    df_testing[Df.RESULT] = pd.Series(result, dtype="float")
    df = calc_gradient_results(df_testing, Df.DATASTREAM_ID)
    pdt.assert_series_equal(df[Df.GRADIENT], pd.Series(ref, name=Df.GRADIENT))


@pytest.mark.parametrize("result", [(range(0, LENGTH, 1)), (range(LENGTH, 0, -1))])
def test_qc_gradient_cacl_vardt(df_testing, result):
    for ds_i in df_testing.datastream_id.unique():
        df_slice = df_testing.loc[df_testing.datastream_id == ds_i]
        df_slice.loc[:, Df.TIME] = pd.Timestamp("now") + pd.timedelta_range(
            start=0, periods=df_slice.shape[0], freq="s", unit="s"  # type: ignore
        ) * list(range(df_slice.shape[0]))
        df_slice.loc[:, Df.RESULT] = pd.Series(result, dtype="float")
        df = calc_gradient_results(df_slice, Df.DATASTREAM_ID)
        pdt.assert_series_equal(
            df[Df.GRADIENT],
            pd.Series(
                np.gradient(
                    df_slice.result, [(1 * i**2) for i in range(df_slice.shape[0])]
                ),
                name=Df.GRADIENT,
            ),
            check_index=False,
        )


@pytest.mark.parametrize(
    "result", [(range(0, LENGTH_SLICE, 1)), (range(LENGTH_SLICE, 0, -1))]
)
def test_qc_gradient_cacl_vardx(df_testing, result):
    def grad_cte_dt(fm1, fp1, dh):
        return (fp1 - fm1) / (2.0 * dh)

    for ds_i in df_testing.datastream_id.unique():
        df_slice = df_testing.loc[df_testing.datastream_id == ds_i]
        df_slice.loc[:, Df.TIME] = pd.Timestamp("now") + pd.timedelta_range(
            start=0, periods=df_slice.shape[0], freq="s", unit="s"  # type: ignore
        )

        df_slice.loc[:, Df.RESULT] = pd.Series(
            np.array(result, "int") * range(df_slice.shape[0]),
            dtype="float",
        )
        df = calc_gradient_results(df_slice, Df.DATASTREAM_ID)

        grad_ref = grad_cte_dt(
            df_slice.result.shift().interpolate(
                method="slinear", limit_direction="both", fill_value="extrapolate"
            ),
            df_slice.result.shift(-1).interpolate(
                method="slinear", limit_direction="both", fill_value="extrapolate"
            ),
            dh=1.0,
        )

        pdt.assert_series_equal(
            df.gradient,
            grad_ref,
            check_index=False,
            check_names=False,
        )


def test_example_pivot_and_reverse():
    df = pd.DataFrame(
        {
            "type": [0, 0, 1, 1],
            "time": [1.1, 2.2, 1.1, 2.2],
            Df.RESULT: list(range(4)),
            "flag": [str(i) for i in range(4)],
        }
    )
    df_p = df.pivot(index=["time"], columns=["type"], values=[Df.RESULT, "flag"])
    df_p_undone = (
        df_p.stack(future_stack=True)  # type: ignore
        .reset_index()
        .sort_values("type")
        .reset_index(drop=True)
        .sort_index(axis=1)
    )
    df_p_undone.result = df_p_undone.result.astype(int)
    pdt.assert_frame_equal(df_p_undone.sort_index(axis=1), df.sort_index(axis=1))


@pytest.mark.parametrize("n", tuple(range(len(base_list_region))))
def test_qc_dependent_quantities(df_testing, n):
    # setup ref count
    qc_flag_count_ref = {
        QualityFlags.GOOD: df_testing.shape[0] - 2,
        QualityFlags.BAD: 2,
    }

    # setup df
    df_testing[Df.QC_FLAG] = QualityFlags.GOOD

    idx_ = df_testing.loc[df_testing[Df.DATASTREAM_ID] == 0].index[n]
    df_testing.loc[idx_, Df.QC_FLAG] = QualityFlags.BAD

    # perform qc check
    qc_flag_update = qc_dependent_quantity_base(
        df_testing, independent=0, dependent=1, dt_tolerance="0.5s"
    )
    df_testing = df_testing.set_index(Df.IOT_ID)
    df_testing.update({Df.QC_FLAG: qc_flag_update})
    assert df_testing[Df.QC_FLAG].value_counts().to_dict() == qc_flag_count_ref


@pytest.mark.parametrize("n", tuple(range(len(base_list_region))))
def test_qc_dependent_quantities_duplicates(df_testing, n):
    # setup ref count
    qc_flag_count_ref = {
        QualityFlags.GOOD: df_testing.shape[0] - 2,
        QualityFlags.BAD: 3,
    }

    # setup df
    df_testing[Df.QC_FLAG] = QualityFlags.GOOD

    idx_ = df_testing.loc[df_testing[Df.DATASTREAM_ID] == 0].index[n]
    df_testing.loc[idx_, Df.QC_FLAG] = QualityFlags.BAD

    idx_duplicate = df_testing.loc[df_testing[Df.DATASTREAM_ID] == 1].index[n]
    df_testing = pd.concat([df_testing, df_testing[df_testing.index==idx_duplicate]], ignore_index=True)

    df_testing.loc[df_testing.index.max(), Df.IOT_ID] = df_testing.loc[df_testing.index.max(), Df.IOT_ID] + df_testing[Df.IOT_ID].max()
    df_testing = df_testing.sort_values(Df.TIME).reset_index(drop=True)

    # perform qc check
    qc_flag_update = qc_dependent_quantity_base(
        df_testing, independent=0, dependent=1, dt_tolerance="0.5s"
    )
    df_testing = df_testing.set_index(Df.IOT_ID)
    df_testing.update({Df.QC_FLAG: qc_flag_update}) # type: ignore
    assert df_testing[Df.QC_FLAG].value_counts().to_dict() == qc_flag_count_ref



@pytest.mark.parametrize("n", tuple(range(len(base_list_region))))
@pytest.mark.parametrize("m", tuple(range(1, len(base_list_region))))
def test_qc_dependent_quantities_missing_dependent(df_testing, n, m):
    # setup ref count
    qc_flag_count_ref = {
        QualityFlags.GOOD: df_testing.shape[0] - 3,
        QualityFlags.BAD: 2,
    }

    # setup df
    df_testing[Df.QC_FLAG] = QualityFlags.GOOD

    idx_ = df_testing.loc[df_testing[Df.DATASTREAM_ID] == 0].index[n]
    df_testing.loc[idx_, Df.QC_FLAG] = QualityFlags.BAD
    pool = cycle(range(len(base_list_region)))
    l = next(pool)
    for i in range(n + m):
        l = next(pool)
    idx_remove = df_testing.loc[df_testing[Df.DATASTREAM_ID] == 1].index[l]

    df_testing = df_testing.drop(index=idx_remove).reset_index()

    # perform qc check
    qc_flag_update = qc_dependent_quantity_base(
        df_testing, independent=0, dependent=1, dt_tolerance="0.5s"
    )
    df_testing = df_testing.set_index(Df.IOT_ID)
    df_testing.update({Df.QC_FLAG: qc_flag_update})
    assert df_testing[Df.QC_FLAG].value_counts().to_dict() == qc_flag_count_ref


def test_qc_range(df_testing):
    df_testing.loc[
        df_testing[Df.DATASTREAM_ID] == 0, ["qc_range_min", "qc_range_max"]
    ] = [2.0, 9.2]
    bool_range = get_bool_out_of_range(df=df_testing, qc_on=Df.RESULT, qc_type="range")
    bool_ref = pd.Series(
        [True, False, False, False, True],
        index=df_testing.loc[df_testing[Df.DATASTREAM_ID] == 0].index,
        dtype=bool,
    )
    pdt.assert_series_equal(bool_ref, bool_range, check_names=False)


@pytest.mark.parametrize("zscore", [25])
@pytest.mark.parametrize("rolling_time_window", ["60min"])
def test_qc_outlier(df_outliers, zscore, rolling_time_window):
    df_outliers[Df.QC_FLAG] = QualityFlags(0)
    df = calc_zscore_results(
        df_outliers, groupby=Df.DATASTREAM_ID, rolling_time_window=rolling_time_window
    )
    df[["qc_zscore_min", "qc_zscore_max"]] = [-1 * zscore, zscore]
    # df[["qc_zscore_min", "qc_zscore_max"]] = [-3.5, 3.5]
    bool_zscore = get_bool_out_of_range(df=df, qc_on=Df.ZSCORE, qc_type="zscore")
    assert bool_zscore.sum() == 9


@pytest.mark.parametrize("zscore", [25])
@pytest.mark.parametrize("rolling_time_window", ["60min"])
def test_qc_outlier_qcflags(df_outliers, zscore, rolling_time_window):
    df = calc_zscore_results(
        df_outliers, groupby=Df.DATASTREAM_ID, rolling_time_window=rolling_time_window
    )
    df[["qc_zscore_min", "qc_zscore_max"]] = [-1 * zscore, zscore]
    # df[["qc_zscore_min", "qc_zscore_max"]] = [-3.5, 3.5]
    bool_zscore = get_bool_out_of_range(df=df, qc_on=Df.ZSCORE, qc_type="zscore")
    assert bool_zscore.sum() == 4


def test_zscore_flag_exclusion(df_testing):
    df_testing[Df.DATASTREAM_ID] = 0
    df_testing[Df.RESULT] = pd.Series(range(df_testing.shape[0]))
    df_all0 = calc_zscore_results(df_testing, groupby=Df.DATASTREAM_ID)
    assert not df_all0[Df.ZSCORE].isnull().any()
    df_testing.loc[5, Df.QC_FLAG] = QualityFlags(4)
    df_one4 = calc_zscore_results(df_testing, groupby=Df.DATASTREAM_ID)
    assert df_one4[Df.ZSCORE].isnull().sum() == 1
    assert not df_one4[Df.RESULT].isnull().any()
    assert not df_testing[Df.RESULT].isnull().any()


def test_zscore_flag_exclusion2(df_testing):
    df_testing[Df.DATASTREAM_ID] = 0
    df_testing.loc[5, Df.QC_FLAG] = QualityFlags(4)
    df_testing = calc_zscore_results(df_testing, groupby=Df.DATASTREAM_ID)
    assert df_testing[Df.ZSCORE].isnull().sum() == 1
    assert not df_testing[Df.RESULT].isnull().any()
    df_testing[Df.ZSCORE] = calc_zscore_results(df_testing, Df.DATASTREAM_ID, rolling_time_window="60min")[Df.ZSCORE]
    assert not df_testing[Df.RESULT].isnull().any()
    assert df_testing[Df.ZSCORE].isnull().sum() == 1


@pytest.mark.parametrize("n", tuple(range(len(base_list_region))))
def test_qc_dependent_quantities_mismatch(df_testing, n):
    # setup ref count
    qc_flag_count_ref = {
        QualityFlags.GOOD: df_testing.shape[0] - 1,
        QualityFlags.BAD: 1,
    }

    # setup df
    df_testing[Df.QC_FLAG] = QualityFlags.GOOD

    idx_ = df_testing.loc[df_testing[Df.DATASTREAM_ID] == 0].index[n]
    df_testing.loc[idx_, Df.TIME] += pd.Timedelta("1d")

    # perform qc check
    qc_update = qc_dependent_quantity_base(
        df_testing, independent=0, dependent=1, dt_tolerance="0.5s"
    )
    df_testing = df_testing.set_index(Df.IOT_ID)
    df_testing.update({Df.QC_FLAG: qc_update})
    assert df_testing[Df.QC_FLAG].value_counts().to_dict() == qc_flag_count_ref


@pytest.mark.parametrize("n", tuple(range(len(base_list_region))))
def test_qc_dependent_quantities_base_3streams(df_testing, n):
    df_additional = df_testing.loc[df_testing[Df.DATASTREAM_ID] == 1]
    df_additional.loc[:, Df.DATASTREAM_ID] = 10
    df_additional.loc[:, Df.IOT_ID] = (
        df_additional[Df.IOT_ID] + df_additional[Df.DATASTREAM_ID]
    )
    df_testing = pd.concat([df_testing, df_additional], ignore_index=True)
    df_testing = df_type_conversions(df_testing)

    df_testing[Df.QC_FLAG] = QualityFlags.GOOD

    idx_ = df_testing.loc[df_testing[Df.DATASTREAM_ID] == 0].index[n]
    df_testing.loc[idx_, Df.QC_FLAG] = QualityFlags.BAD

    qc_flag_count_ref = {
        QualityFlags.GOOD: df_testing.shape[0] - 2,
        QualityFlags.BAD: 2,
    }
    qc_update = qc_dependent_quantity_base(
        df_testing, independent=0, dependent=1, dt_tolerance="0.5s"
    )
    df_testing = df_testing.set_index(Df.IOT_ID)
    df_testing.update({Df.QC_FLAG: qc_update})
    assert df_testing[Df.QC_FLAG].value_counts().to_dict() == qc_flag_count_ref


@pytest.mark.parametrize("n", tuple(range(len(base_list_region))))
@pytest.mark.parametrize("n_rel_del", tuple(range(1, len(base_list_region))))
@pytest.mark.parametrize(
    "independent_id, dependent_id",
    [(0, 1), (1, 0)],
)
def test_qc_dependent_quantities_base_3streams_missing(
    df_testing, n, n_rel_del, independent_id, dependent_id
):
    df_additional = df_testing.loc[df_testing[Df.DATASTREAM_ID] == dependent_id]
    df_additional.loc[:, Df.DATASTREAM_ID] = 10
    df_additional.loc[:, Df.IOT_ID] = (
        df_additional[Df.IOT_ID] + df_additional[Df.DATASTREAM_ID]
    )
    df_testing = pd.concat([df_testing, df_additional], ignore_index=True)
    df_testing = df_type_conversions(df_testing)

    df_testing[Df.QC_FLAG] = QualityFlags.GOOD

    idx_ = df_testing.loc[df_testing[Df.DATASTREAM_ID] == independent_id].index[n]
    idx_delete = next(
        islice(
            cycle(
                df_testing.loc[
                    df_testing[Df.DATASTREAM_ID] == independent_id
                ].index.values
            ),
            n + n_rel_del,
            None,
        )
    )
    df_testing.loc[idx_, Df.QC_FLAG] = QualityFlags.BAD
    df_testing = df_testing.drop(idx_delete).reset_index()

    qc_flag_count_ref = {
        QualityFlags.GOOD: df_testing.shape[0] - 3,
        QualityFlags.BAD: 3,
    }
    qc_update = qc_dependent_quantity_base(
        df_testing,
        independent=independent_id,
        dependent=dependent_id,
        dt_tolerance="0.5s",
        flag_when_missing=QualityFlags.BAD,
    )
    df_testing = df_testing.set_index(Df.IOT_ID)
    df_testing.update({Df.QC_FLAG: qc_update})
    assert df_testing[Df.QC_FLAG].value_counts().to_dict() == qc_flag_count_ref


@pytest.mark.parametrize("n", tuple(range(len(base_list_region))))
@pytest.mark.parametrize("n_rel_del", tuple(range(1, len(base_list_region))))
@pytest.mark.parametrize(
    "independent_id, dependent_id",
    [(0, 1), (1, 0)],
)
def test_qc_dependent_quantities_base_3streams_missing_dependent(
    df_testing, n, n_rel_del, independent_id, dependent_id
):
    df_additional = df_testing.loc[df_testing[Df.DATASTREAM_ID] == dependent_id]
    df_additional.loc[:, Df.DATASTREAM_ID] = 10
    df_additional.loc[:, Df.IOT_ID] = (
        df_additional[Df.IOT_ID] + df_additional[Df.DATASTREAM_ID]
    )
    df_testing = pd.concat([df_testing, df_additional], ignore_index=True)
    df_testing = df_type_conversions(df_testing)

    df_testing[Df.QC_FLAG] = QualityFlags.GOOD

    idx_ = df_testing.loc[df_testing[Df.DATASTREAM_ID] == independent_id].index[n]
    idx_delete = next(
        islice(
            cycle(
                df_testing.loc[
                    df_testing[Df.DATASTREAM_ID] == dependent_id
                ].index.values
            ),
            n + n_rel_del,
            None,
        )
    )
    df_testing.loc[idx_, Df.QC_FLAG] = QualityFlags.BAD
    df_testing = df_testing.drop(idx_delete).reset_index()

    qc_flag_count_ref = {
        QualityFlags.GOOD: df_testing.shape[0] - 2,
        QualityFlags.BAD: 2,
    }
    qc_update = qc_dependent_quantity_base(
        df_testing,
        independent=independent_id,
        dependent=dependent_id,
        dt_tolerance="0.5s",
        flag_when_missing=QualityFlags.BAD,
    )
    df_testing = df_testing.set_index(Df.IOT_ID)
    df_testing.update({Df.QC_FLAG: qc_update})
    assert df_testing[Df.QC_FLAG].value_counts().to_dict() == qc_flag_count_ref


@pytest.mark.parametrize("n", tuple(range(len(base_list_region))))
@pytest.mark.parametrize("n_rel_del", tuple(range(1, len(base_list_region))))
@pytest.mark.parametrize(
    "independent_id, dependent_id",
    [(0, 1), (1, 0)],
)
def test_qc_dependent_quantities_base_3streams_missing_noflag(
    df_testing, n, n_rel_del, independent_id, dependent_id
):
    df_additional = df_testing.loc[df_testing[Df.DATASTREAM_ID] == dependent_id]
    df_additional.loc[:, Df.DATASTREAM_ID] = 10
    df_additional.loc[:, Df.IOT_ID] = (
        df_additional[Df.IOT_ID] + df_additional[Df.DATASTREAM_ID]
    )
    df_testing = pd.concat([df_testing, df_additional], ignore_index=True)
    df_testing = df_type_conversions(df_testing)

    df_testing[Df.QC_FLAG] = QualityFlags.GOOD

    idx_ = df_testing.loc[df_testing[Df.DATASTREAM_ID] == independent_id].index[n]
    idx_delete = next(
        islice(
            cycle(
                df_testing.loc[
                    df_testing[Df.DATASTREAM_ID] == independent_id
                ].index.values
            ),
            n + n_rel_del,
            None,
        )
    )
    df_testing.loc[idx_, Df.QC_FLAG] = QualityFlags.BAD
    df_testing = df_testing.drop(idx_delete).reset_index()

    qc_flag_count_ref = {
        QualityFlags.GOOD: df_testing.shape[0] - 2,
        QualityFlags.BAD: 2,
    }
    qc_update = qc_dependent_quantity_base(
        df_testing,
        independent=independent_id,
        dependent=dependent_id,
        dt_tolerance="0.5s",
        flag_when_missing=None,
    )
    df_testing = df_testing.set_index(Df.IOT_ID)
    df_testing.update(qc_update)
    assert df_testing[Df.QC_FLAG].value_counts().to_dict() == qc_flag_count_ref


@pytest.mark.parametrize("bad_value", (100.0,))
@pytest.mark.parametrize("n", (0, 2, 4))
@pytest.mark.parametrize(
    "independent_id, dependent_id",
    [
        (0, 1),
    ],
)
def test_qc_dependent_quantities_secondary_fct(
    df_testing, bad_value, n, independent_id, dependent_id
):
    qc_flag_count_ref = {
        QualityFlags.GOOD: df_testing.shape[0] - 1,
        QualityFlags.BAD: 1,
    }
    df_testing[Df.QC_FLAG] = QualityFlags.GOOD

    idx_ = df_testing[df_testing[Df.DATASTREAM_ID] == independent_id].index[n]
    df_testing.loc[idx_, Df.RESULT] = bad_value

    qc_update = qc_dependent_quantity_secondary(
        df_testing,
        independent=independent_id,
        dependent=dependent_id,
        range_=(0.0, 10.0),
        dt_tolerance="0.5s",
    )
    df_testing = df_testing.set_index(Df.IOT_ID)
    df_testing.update(qc_update)

    assert df_testing[Df.QC_FLAG].value_counts().to_dict() == qc_flag_count_ref
    assert (
        df_testing.loc[
            idx_ + int(df_testing.shape[0] / len(df_testing.datastream_id.unique())),
            Df.QC_FLAG,
        ]
        == QualityFlags.BAD
    )


@pytest.mark.parametrize("bad_value", (100.0,))
@pytest.mark.parametrize("n", (0, 2, 4))
@pytest.mark.parametrize("m", (1, 3))
@pytest.mark.parametrize(
    "independent_id, dependent_id",
    [
        (0, 1),
    ],
)
def test_qc_dependent_quantities_secondary_fct_missing_dependent(
    df_testing, bad_value, n, m, independent_id, dependent_id
):
    qc_flag_count_ref = {
        QualityFlags.GOOD: df_testing.shape[0] - 2,
        QualityFlags.BAD: 1,
    }
    df_testing[Df.QC_FLAG] = QualityFlags.GOOD

    idx_ = df_testing[df_testing[Df.DATASTREAM_ID] == independent_id].index[n]
    df_testing.loc[idx_, Df.RESULT] = bad_value

    pool = cycle(range(len(base_list_region)))
    l = next(pool)
    for i in range(n + m):
        l = next(pool)
    idx_remove = df_testing.loc[df_testing[Df.DATASTREAM_ID] == independent_id].index[l]

    df_testing = df_testing.drop(index=idx_remove).reset_index()

    qc_update = qc_dependent_quantity_secondary(
        df_testing,
        independent=independent_id,
        dependent=dependent_id,
        range_=(0.0, 10.0),
        dt_tolerance="0.5s",
    )
    df_testing = df_testing.set_index(Df.IOT_ID)
    df_testing.update({Df.QC_FLAG: qc_update})

    assert df_testing[Df.QC_FLAG].value_counts().to_dict() == qc_flag_count_ref
    assert (
        df_testing.loc[
            idx_
            + int(
                (df_testing.shape[0] + 1) / len(df_testing[Df.DATASTREAM_ID].unique())
            ),  # TODO: refactoring!
            Df.QC_FLAG,
        ]
        == QualityFlags.BAD
    )

    def test_combine_dicts(self):
        out = combine_dicts(
            {"first": 1, "str": "test", "float": 2.3},
            {"second": 2, "str": "ing", "float": 4.5},
        )
        assert out == {"first": 1, "str": "testing", "second": 2, "float": 6.8}
